@@ -1,5 +1,11 @@
+from typing import Dict, Generator, List, Optional, Set, Tuple, Type, Union
+
 import numpy as np
 
+from .cartesian_graph import CartesianGraph
+from .node import Node, OperatorNode
+from .primitives import Primitives
+
 try:
     import torch  # noqa: F401
 
@@ -7,12 +13,6 @@ try:
 except ModuleNotFoundError:
     torch_available = False
 
-from typing import Dict, Generator, List, Optional, Tuple, Type
-
-from .cartesian_graph import CartesianGraph
-from .node import Node, Parameter
-from .primitives import Primitives
-
 
 ID_INPUT_NODE: int = -1
 ID_OUTPUT_NODE: int = -2
@@ -29,7 +29,7 @@ class Genome:
         n_outputs: int,
         n_columns: int,
         n_rows: int,
-        levels_back: int,
+        levels_back: Union[int, None],
         primitives: Tuple[Type[Node], ...],
     ) -> None:
         """Init function.
@@ -44,11 +44,12 @@ class Genome:
             Number of columns in the representation of the genome.
         n_rows : int
             Number of rows in the representation of the genome.
-        levels_back : int
-            Number of previous columns that an entry in the genome can be
-            connected with.
+        levels_back : Union[int, None]
+            Maximal column distance of inputs to an internal node. If
+            set to `None`, no restrictions are used.
         primitives : Tuple[Type[Node], ...]
            Tuple of primitives that the genome can refer to.
+
         """
         if n_inputs <= 0:
             raise ValueError("n_inputs must be strictly positive")
@@ -89,6 +90,9 @@ class Genome:
         # dictionary to store values of Parameter nodes
         self._parameter_names_to_values: Dict[str, float] = {}
 
+        # list of permissible values for every gene
+        self._permissible_values: List[np.ndarray] = self.determine_permissible_values()
+
     def __getitem__(self, key: int) -> int:
         if self.dna is None:
             raise RuntimeError("dna not initialized")
@@ -107,7 +111,7 @@ class Genome:
     def dna(self, value: List[int]) -> None:
         self._validate_dna(value)
         self._dna = value
-        self._initialize_unkown_parameters()
+        self._initialize_unknown_parameters()
 
     @property
     def _n_hidden(self) -> int:
@@ -133,6 +137,70 @@ class Genome:
         s += ")"
         return s
 
+    def determine_permissible_values_per_gene(self, gene_idx: int) -> np.ndarray:
+        region_idx = self._get_region_idx(gene_idx)
+
+        if self._is_input_region(region_idx):
+            return self._determine_permissible_values_input_region(gene_idx)
+
+        elif self._is_hidden_region(region_idx):
+            return self._determine_permissible_values_hidden_region(gene_idx, region_idx)
+
+        elif self._is_output_region(region_idx):
+            return self._determine_permissible_values_output_region(gene_idx)
+
+        else:
+            assert False  # should never be reached
+
+    def determine_permissible_values(self) -> List[np.ndarray]:
+        """Determine permissible values for every gene.
+
+        Parameters
+        ----------
+        None
+
+        Returns
+        ----------
+        permissible_values
+            List[numpy.ndarray]: List of permissible values for every gene
+        """
+        permissible_values: List[np.ndarray] = []
+        for gene_idx in range(self._n_genes):
+            permissible_values_per_gene = self.determine_permissible_values_per_gene(gene_idx)
+            permissible_values.append(permissible_values_per_gene)
+        return permissible_values
+
+    def _determine_permissible_values_input_region(self, gene_idx: int) -> np.ndarray:
+
+        if self._is_function_gene(gene_idx):
+            return np.array(self._id_input_node)
+        else:
+            return np.array(self._id_unused_gene)
+
+    def _determine_permissible_values_hidden_region(
+        self, gene_idx: int, region_idx: int
+    ) -> np.ndarray:
+
+        if self._is_function_gene(gene_idx):
+            return np.arange(len(self._primitives._primitives))
+
+        elif self._is_hidden_input_gene(gene_idx, region_idx):
+            return np.array(self._permissible_inputs(region_idx))
+
+        else:
+            assert False  # should never be reached
+
+    def _determine_permissible_values_output_region(self, gene_idx: int) -> np.ndarray:
+
+        if self._is_function_gene(gene_idx):
+            return np.array(self._id_output_node)
+        else:
+            input_index = gene_idx % self._length_per_region - 1
+            if input_index == 0:
+                return np.array(self._permissible_inputs_for_output_region())
+            else:
+                return np.array(self._id_unused_gene)
+
     def _create_input_region(self) -> List[int]:
 
         region = []
@@ -171,7 +239,7 @@ class Genome:
         Parameters
         ----------
         rng : numpy.RandomState
-            Random number generator instance to use for crossover.
+            Random number generator instance to use for randomizing.
 
         Returns
         ----------
@@ -199,6 +267,147 @@ class Genome:
         # accept generated dna if it is valid
         self.dna = dna
 
+    def reorder(self, rng: np.random.RandomState) -> None:
+        """Reorder the genome
+
+        Shuffle node ordering of internal (hidden) nodes in genome without changing the phenotype.
+        (Goldman 2015, DOI: 10.1109/TEVC.2014.2324539)
+
+        During reordering, inactive genes, e.g., input genes of nodes with arity zero,
+        are not taken into account and can hence have invalid values after reordering.
+        These invalid values are replaced by random values
+        for the respective gene after reordering.
+
+        Parameters
+        ----------
+        rng : numpy.RandomState
+            Random number generator instance.
+
+        Returns
+        ----------
+        None
+        """
+        if (self._n_rows != 1) or (self._levels_back != self._n_columns):
+            raise ValueError(
+                "Genome reordering is only implemented for n_rows=1" " and levels_back=n_columns"
+            )
+
+        dna = self._dna.copy()
+
+        node_dependencies: Dict[int, Set[int]] = self._determine_node_dependencies()
+
+        addable_nodes: Set[int] = self._get_addable_nodes(node_dependencies)
+
+        new_node_idx: int = self._n_inputs  # First position to be placed is after inputs
+        used_node_indices: List[int] = []
+
+        while len(addable_nodes) > 0:
+
+            old_node_idx = rng.choice(list(addable_nodes))
+
+            dna = self._copy_dna_segment(dna, old_node_idx=old_node_idx, new_node_idx=new_node_idx)
+
+            for dependencies in node_dependencies.values():
+                dependencies.discard(old_node_idx)
+
+            used_node_indices.append(old_node_idx)
+            addable_nodes = self._get_addable_nodes(node_dependencies, used_node_indices)
+            new_node_idx += 1
+
+        self._update_input_genes(dna, used_node_indices)
+        self._replace_invalid_input_alleles(dna, rng)
+
+        self.dna = dna
+
+    def _copy_dna_segment(self, dna: List[int], old_node_idx: int, new_node_idx: int) -> List[int]:
+        """ Copy a nodes dna segment from its old node location to a new location. """
+
+        dna[
+            new_node_idx * self._length_per_region : (new_node_idx + 1) * self._length_per_region
+        ] = self._dna[
+            old_node_idx * self._length_per_region : (old_node_idx + 1) * self._length_per_region
+        ]
+
+        return dna
+
+    def _update_input_genes(self, dna: List[int], used_node_indices: List[int]) -> None:
+        """Update input genes of all nodes from old node indices to new node indices"""
+        for gene_idx, gene_value in enumerate(dna):
+            region_idx = self._get_region_idx(gene_idx)
+            if self._is_hidden_input_gene(gene_idx, region_idx) or self._is_output_input_gene(
+                gene_idx
+            ):
+                if gene_value >= self._n_inputs:
+                    gene_value = self._n_inputs + used_node_indices.index(gene_value)
+            dna[gene_idx] = gene_value
+
+    def _replace_invalid_input_alleles(self, dna: List[int], rng: np.random.RandomState) -> None:
+        """Replace invalid alleles for unused input genes of all nodes
+        by random permissible values.
+        WARNING: Works only if self.n_rows==1.
+        """
+        assert self._n_rows == 1
+
+        for gene_idx, gene_value in enumerate(dna):
+            region_idx = self._get_region_idx(gene_idx)
+            if self._is_hidden_input_gene(gene_idx, region_idx) and gene_value > region_idx:
+                permissible_values = self.determine_permissible_values_per_gene(gene_idx)
+                gene_value = rng.choice(permissible_values)
+                dna[gene_idx] = gene_value
+
+    def _get_addable_nodes(
+        self, node_dependencies: Dict[int, Set[int]], used_node_indices: List[int] = [],
+    ) -> Set[int]:
+        """ Get the set of addable nodes,
+         nodes which have no dependencies and were not already used.
+        """
+        addable_nodes = set(
+            idx for idx, dependencies in node_dependencies.items() if len(dependencies) == 0
+        )
+        return addable_nodes.difference(used_node_indices)
+
+    def _get_region_idx(self, gene_idx: int) -> int:
+
+        return gene_idx // self._length_per_region
+
+    def _determine_node_dependencies(self) -> Dict[int, Set[int]]:
+        """ Determines the set of node indices on which each node depends.
+            Unused input genes are ignored.
+
+        Returns
+        ----
+        dependencies: Dict[int, Set[int]]
+            Dictionary containing for every node the set of active input genes
+
+        """
+        dependencies: Dict[int, Set[int]] = {}
+        for region_idx, _ in self.iter_hidden_regions():
+
+            current_node_dependencies: Set[int] = set()
+
+            operator_idx: int = region_idx * self._length_per_region
+
+            current_arity: int = self._determine_operator_arity(operator_idx)
+
+            for idx_gene in range(
+                1, current_arity + 1
+            ):  # shift by 1 since first gene is the operator gene
+                input_node_idx = self._dna[operator_idx + idx_gene]
+                if not self._is_input_region(
+                    input_node_idx
+                ):  # not necessary to add input regions, since their positions remain fixed
+                    current_node_dependencies.add(input_node_idx)
+
+            dependencies[region_idx] = current_node_dependencies
+
+        return dependencies
+
+    def _determine_operator_arity(self, gene_idx: int) -> int:
+
+        assert self._is_function_gene(gene_idx)
+
+        return self._primitives[self._dna[gene_idx]]._arity
+
     def _permissible_inputs(self, region_idx: int) -> List[int]:
 
         assert not self._is_input_region(region_idx)
@@ -313,7 +522,7 @@ class Genome:
         return gene_idx < (self._n_inputs * self._length_per_region)
 
     def _is_gene_in_hidden_region(self, gene_idx: int) -> bool:
-        return ((self._n_inputs * self._length_per_region) <= gene_idx) & (
+        return ((self._n_inputs * self._length_per_region) <= gene_idx) and (
             gene_idx < ((self._n_inputs + self._n_hidden) * self._length_per_region)
         )
 
@@ -324,7 +533,7 @@ class Genome:
         return region_idx < self._n_inputs
 
     def _is_hidden_region(self, region_idx: int) -> bool:
-        return (self._n_inputs <= region_idx) & (region_idx < self._n_inputs + self._n_hidden)
+        return (self._n_inputs <= region_idx) and (region_idx < self._n_inputs + self._n_hidden)
 
     def _is_output_region(self, region_idx: int) -> bool:
         return self._n_inputs + self._n_hidden <= region_idx
@@ -332,19 +541,34 @@ class Genome:
     def _is_function_gene(self, gene_idx: int) -> bool:
         return (gene_idx % self._length_per_region) == 0
 
-    def _is_active_input_gene(self, gene_idx: int) -> bool:
-        input_index = gene_idx % self._length_per_region
-        assert input_index > 0
-        region_idx = gene_idx // self._length_per_region
-        if self._is_input_region(region_idx):
-            return False
-        elif self._is_hidden_region(region_idx):
-            node_arity = self._primitives[self.dna[region_idx * self._length_per_region]]._arity
-            return input_index <= node_arity
-        elif self._is_output_region(region_idx):
-            return input_index == 1
-        else:
-            assert False  # should never be reached
+    def _is_hidden_input_gene(self, gene_idx: int, region_idx: int) -> bool:
+        return self._is_hidden_region(region_idx) and (not self._is_function_gene(gene_idx))
+
+    def _is_output_input_gene(self, gene_idx: int) -> bool:
+        return (
+            self._is_gene_in_output_region(gene_idx) and gene_idx % self._length_per_region == 1
+        )  # assumes 2nd gene in output is coding for input
+
+    def _select_gene_indices_for_mutation(
+        self, mutation_rate: float, rng: np.random.RandomState
+    ) -> List[int]:
+        """Selects the gene indices for mutations
+
+        Parameters
+        ----------
+        mutation_rate : float
+            Probability of a gene to be mutated, between 0 (excluded) and 1 (included).
+        rng : numpy.random.RandomState
+            Random number generator instance to use for selecting the indices.
+
+        Returns
+        ----------
+        selected_gene_indices: np.ndarray
+            indices of the genes selected for mutation.
+        """
+
+        selected_gene_indices = np.nonzero(rng.rand(len(self.dna)) < mutation_rate)[0]
+        return selected_gene_indices
 
     def mutate(self, mutation_rate: float, rng: np.random.RandomState):
         """Mutate the genome.
@@ -352,9 +576,10 @@ class Genome:
         Parameters
         ----------
         mutation_rate : float
-            Proportion of genes to be mutated, between 0 and 1.
+            Probability of a gene to be mutated, between 0 (excluded) and 1 (included).
         rng : numpy.random.RandomState
-            Random number generator instance to use for crossover.
+            Random number generator instance to use for mutating.
+
 
         Returns
         ----------
@@ -362,72 +587,30 @@ class Genome:
             True if only inactive regions of the genome were mutated, False otherwise.
         """
 
-        def count_dna_differences(dna0: List[int], dna1: List[int]) -> int:
-            return len([1 for gene0, gene1 in zip(dna0, dna1) if gene0 != gene1])
-
-        n_mutations = int(mutation_rate * len(self.dna))
-        assert n_mutations > 0
-
         graph = CartesianGraph(self)
         active_regions = graph.determine_active_regions()
-
         dna = list(self._dna)
-
         only_silent_mutations = True
-        while count_dna_differences(self.dna, dna) < n_mutations:
 
-            gene_idx = rng.randint(0, self._n_genes)
-            region_idx = gene_idx // self._length_per_region
+        selected_gene_indices = self._select_gene_indices_for_mutation(mutation_rate, rng)
 
-            if self._is_input_region(region_idx):
-                continue  # nothing to do here
+        for (gene_idx, allele) in zip(selected_gene_indices, np.array(dna)[selected_gene_indices]):
 
-            elif self._is_output_region(region_idx):
-                silent = self._mutate_output_region(dna, gene_idx, rng)
+            region_idx = self._get_region_idx(gene_idx)
 
-            elif self._is_hidden_region(region_idx):
-                silent = self._mutate_hidden_region(dna, gene_idx, active_regions, rng)
+            permissible_values = self._permissible_values[gene_idx]
+            permissible_alternative_values = permissible_values[permissible_values != allele]
 
-            else:
-                assert False  # should never be reached
+            if len(permissible_alternative_values) > 0:
 
-            only_silent_mutations = only_silent_mutations and silent
+                dna[gene_idx] = rng.choice(permissible_alternative_values)
+                silent = region_idx not in active_regions
+
+                only_silent_mutations = only_silent_mutations and silent
 
         self.dna = dna
         return only_silent_mutations
 
-    def _mutate_output_region(
-        self, dna: List[int], gene_idx: int, rng: np.random.RandomState
-    ) -> bool:
-        assert self._is_gene_in_output_region(gene_idx)
-
-        if not self._is_function_gene(gene_idx) and self._is_active_input_gene(gene_idx):
-            permissible_inputs = self._permissible_inputs_for_output_region()
-            dna[gene_idx] = rng.choice(permissible_inputs)
-            return False
-        else:
-            return True
-
-    def _mutate_hidden_region(
-        self, dna: List[int], gene_idx: int, active_regions: List[int], rng: np.random.RandomState
-    ) -> bool:
-
-        assert self._is_gene_in_hidden_region(gene_idx)
-
-        region_idx = gene_idx // self._length_per_region
-        silent_mutation = region_idx not in active_regions
-
-        if self._is_function_gene(gene_idx):
-            dna[gene_idx] = self._primitives.sample_allele(rng)
-            return silent_mutation
-
-        else:
-            permissible_inputs = self._permissible_inputs(region_idx)
-            dna[gene_idx] = rng.choice(permissible_inputs)
-
-            silent_mutation = silent_mutation or (not self._is_active_input_gene(gene_idx))
-            return silent_mutation
-
     @property
     def primitives(self) -> Primitives:
         return self._primitives
@@ -484,13 +667,14 @@ class Genome:
 
         return any_parameter_updated
 
-    def _initialize_unkown_parameters(self) -> None:
+    def _initialize_unknown_parameters(self) -> None:
         for region_idx, region in self.iter_hidden_regions():
             node_id = region[0]
             node_type = self._primitives[node_id]
-            parameter_name = f"<p{region_idx}>"
-            if (
-                issubclass(node_type, Parameter)
-                and parameter_name not in self._parameter_names_to_values
-            ):
-                self._parameter_names_to_values[parameter_name] = node_type.initial_value()
+            assert issubclass(node_type, OperatorNode)
+            for parameter_name in node_type._parameter_names:
+                parameter_name_with_idx = "<" + parameter_name[1:-1] + str(region_idx) + ">"
+                if parameter_name_with_idx not in self._parameter_names_to_values:
+                    self._parameter_names_to_values[
+                        parameter_name_with_idx
+                    ] = node_type.initial_value(parameter_name_with_idx)
